-- EthOS - Spirit System Script
-- Copyright (C) 2024 Spirit System, email: dev@spirit-system.com
-- This file is distributed under Creative Commons 4 license - BY-NC-ND
-- Script version: 3.7.1-r2

local translations = {en="Spirit FBL"}

local function name(widget)
  local locale = system.getLocale()
  return translations[locale] or translations["en"]
end

local requestExit = false
local requestInProgress = false
local refreshIndex = 0
local modifications = {}
local fields = {}

local parameters = {
  -- { name, type, default, value, min, max }
  {"", createMenuButton, nil, pageGeneral, "Base"},
}

local logentries = {
  -- { name, type, default, value, min, max }
}

local logfields = {}

local pagePrev
local pageMenu
local pageGeneral
local pageSensor
local pageStabi
local pageLimits
local pageAdvanced
local pageExpert

local skipField = 0

local requestBank = false
local requestBankInProgress = false
local banksEnabled = false
local bankNumber = 0
local requestLogInProgress = false

local cyclic_type
local rudder_type
local cyclic_freq
local rudder_freq

local log_duration
local log_loaded

function bitoper(a, b)
  return a & b
end

local function getValue(parameter)
  if parameter[4] == 0x50+3 then
    bankNumber = parameter[5]
  elseif parameter[4] == 0x50+6 then
    cyclic_type = parameter[5]
  elseif parameter[4] == 0x50+8 then
    rudder_type = parameter[5]
  elseif parameter[4] == 0x50+7 then
    cyclic_freq = parameter[5]
  elseif parameter[4] == 0x50+9 then
    rudder_freq = parameter[5]
  elseif parameter[4] == 0x50+171 then
    log_duration = parameter[5]
  elseif parameter[4] == 0x50+173 then
    return 0
  end
  
  --if banksEnabled and parameter[]
  if parameter[5] == nil then
    return 0
  else
    return parameter[5]
  end
end


  
local function exitQuestion()  
  local buttons = {
    {label="Yes", action=function () modifications[#modifications+1] = {0x50+161, 0}; return true end},
    {label="No", action=function() return true end},
  }
  
  form.openDialog("Spirit FBL", "Do you want to close the Menu?", buttons)
end
    
local function setValue(parameter, value)  
  -- Servo Frequency and Center Pulse verification
  if parameter[4] == 0x50+6 and cyclic_freq > 70 and value ~= 66 then
    value = 66
  elseif parameter[4] == 0x50+8 and rudder_freq > 70 and value ~= 67 then
    value = 67
  elseif parameter[4] == 0x50+9 and value > 70 and rudder_type ~= 67 then
    value = 65
  elseif parameter[4] == 0x50+7 and value > 70 and cyclic_type ~= 66 then
    value = 65
  end
  
  if parameter[4] == 0x50+161 and value == 0 then
    if requestBankInProgress == true then
      requestExit = true
    else
      exitQuestion()
      return true
    end
  else
    parameter[5] = value
    modifications[#modifications+1] = {parameter[4], value}
  end
end

local function getLog()
  requestLogInProgress = false
end

local function updateParameters(param)
  fields = {}
  refreshIndex = 0
  form.clear()
  requestInProgress = false
  skipField = 0
  requestBank = true
  requestLogInProgress = false
  
  log_duration = 0
  logentries = {}
  logfields = {}
  
  if param == nil then
    parameters = pagePrev
  else
    pagePrev = parameters
    parameters = param
  end
    
  for index = 1, #parameters do
    local parameter = parameters[index]
    local line = form.addLine(parameter[2])
    local field = parameter[3](line, parameter)
    fields[#fields + 1] = field
  end
end

local function createNumberField(line, parameter)
  local offset = parameter[9]
  if #parameter < 9 then
    offset = 0
  end
  
  local field = form.addNumberField(line, nil, parameter[6], parameter[7], function() return (getValue(parameter)+offset) end, function(value) setValue(parameter, value-offset) end)
  field:enableInstantChange(false)
  if #parameter >= 8 and parameter[8] ~= nil then
    field:suffix(parameter[8])
  end
  
  if parameter[4] == 0x50+24 or parameter[4] == 0x50+26 or parameter[4] == 0x50+73 or parameter[4] == 0x50+138 or parameter[4] == 0x50+38 then
    field:decimals (1)
  end  
  field:enable(false)
  
  return field
end

local function createChoiceField(line, parameter)
  local field = form.addChoiceField(line, nil, parameter[6], function() return getValue(parameter) end, function(value) setValue(parameter, value) end)
  field:enable(false)
  return field
end

local function createChannelField(line, parameter)
  local field = form.addNumberField(line, nil, parameter[6], parameter[7], function() return getValue(parameter) + 1 end, function(value) setValue(parameter, value - 1) end)
  field:prefix("CH")
  field:enableInstantChange(false)
  field:enable(false)
  return field
end

local function createTextButton(line, parameter)
  local field = form.addTextButton(line, nil, parameter[6], function() return setValue(parameter, 0) end)
  field:enable(true)
  return field
end

local function createMenuButton(line, parameter)
  local field = form.addTextButton(line, nil, parameter[6], function() return updateParameters(parameter[5]) end)
  field:enable(true)
  return field
end

local function createTextLabel(line, parameter)
  local field = form.addStaticText(line, nil, parameter[6])
  --field:enable(true)
  return field
end

pageGeneral = {
  -- { name, type, default, value, min, max }
  {0, "> GENERAL", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "Receiver Type", createChoiceField, 0x50+4, nil, {{"PWM", 65}, {"PPM", 66}, {"S-BUS", 68}, {"F.Port", 72}, {"F.Bus", 74}}},
  {0, "Swashplate Type", createChoiceField, 0x50+5, nil, {{"CCPM 120", 65}, {"CCPM 120 rev", 66}, {"CCPM 135", 67}, {"CCPM 135 rev", 68}, {0, "CCPM 140", 69}, {"CCPM 140 rev", 70}, {"CCPM 90", 71}, {"CCPM 90 Velos", 72}, {"Tandem", 73}}},
  {0, "Position", createChoiceField, 0x50+2, nil, {{"Horz(0)", 65}, {"Horz(180)", 66}, {"Horz(0+i)", 67}, {"Horz(180+i)", 68}, {"Vert(0-l)", 69}, {"Vert(180-l)", 70}, {"Vert(0-r)", 71}, {"Vert(180-r)", 72}}},
  {1, "Flight Style", createNumberField, 0x50+44, nil, 1, 8, nil, 1 },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageServos = {
  -- { name, type, default, value, min, max }
  {0, "> SERVOS", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "CYCLIC", createTextLabel, nil, nil, "" },
  {0, "Center Pulse", createChoiceField, 0x50+6, nil, {{"1520us", 65}, {"760us", 66}}},
  {0, "Servo Frequency", createChoiceField, 0x50+7, nil, {{"50Hz", 65}, {"60Hz", 66}, {"100Hz", 67}, {"150Hz", 68}, {"200Hz", 69}, {"333Hz", 70}, {"560Hz", 71}, {"700Hz", 72}}},
  {0, "", createTextLabel, nil, nil, "" },
  {0, "RUDDER", createTextLabel, nil, nil, "" },
  {0, "Center Pulse", createChoiceField, 0x50+8, nil, {{"1520us", 65}, {"960us", 66}, {"760us", 67}}},
  {0, "Servo Frequency", createChoiceField, 0x50+9, nil, {{"50Hz", 65}, {"60Hz", 66}, {"100Hz", 67}, {"150Hz", 68}, {"200Hz", 69}, {"333Hz", 70}, {"560Hz", 71}, {"700Hz", 72}}},
  {0, "", createTextLabel, nil, nil, "" },
  {0, "SUBTRIM", createTextLabel, nil, nil, "" },
  {0, "CH1", createNumberField, 0x50+15, nil, -127, 127, nil, -128 },
  {0, "CH2", createNumberField, 0x50+16, nil, -127, 127, nil, -128 },
  {0, "CH3", createNumberField, 0x50+17, nil, -127, 127, nil, -128 },
  {0, "CH4", createNumberField, 0x50+11, nil, -127, 127, nil, -128 },
  {0, "", createTextLabel, nil, nil, "" },
  {0, "SERVO REVERSE", createTextLabel, nil, nil, "" },
  {0, "CH1", createChoiceField, 0x50+21, nil, {{"No", 48}, {"Yes", 49}}},
  {0, "CH2", createChoiceField, 0x50+22, nil, {{"No", 48}, {"Yes", 49}}},
  {0, "CH3", createChoiceField, 0x50+23, nil, {{"No", 48}, {"Yes", 49}}},
  {0, "CH4", createChoiceField, 0x50+31, nil, {{"No", 48}, {"Yes", 49}}},
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageSensor = {
  -- { name, type, default, value, min, max }
  {0, "> SENSOR", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {1, "Cyclic Gain", createNumberField, 0x50+18, nil, 25, 100, "%", 20 },
  {1, "Rudder Gain", createNumberField, 0x50+59, nil, -100, 100, "%", -100 },
  {1, "Rudder Common Gain", createNumberField, 0x50+20, nil, 100, 250, "%", 50 },
  {1, "Cyclic Rotation Speed", createNumberField, 0x50+24, nil, 40, 160 },
  {1, "Rudder Rotation Speed", createNumberField, 0x50+26, nil, 40, 200 },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageThrottle = {
  -- { name, type, default, value, min, max }
  {0, "> THROTTLE", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "Reverse", createChoiceField, 0x50+66, nil, {{"No", 48}, {"Yes", 49}}},
  {0, "Frequency", createChoiceField, 0x50+71, nil, {{"60Hz", 65}, {"120Hz", 66}, {"200Hz", 67}, {"333Hz", 68}}},
  {0, "Range - Min.", createNumberField, 0x50+67, nil, 7, 13, "0 us" },
  {0, "Range - Max.", createNumberField, 0x50+68, nil, 17, 23, "0 us" },
  {0, "", createTextLabel, nil, nil, "" },
  {1, "Governor", createChoiceField, 0x50+72, nil, {{"Disabled", 0}, {"Enabled", 1}}},
  {0, "Spoolup Rampup", createChoiceField, 0x50+60, nil, {{"Very slow", 0}, {"Slow", 1}, {"Fast", 2}, {"Very fast", 3}}},
  {0, "Sensing Divider", createNumberField, 0x50+64, nil, 1, 14 },
  {0, "Gear Ratio", createNumberField, 0x50+65, nil, 1, 480, " /20" },
  {0, "Max. Head Speed", createNumberField, 0x50+69, nil, 150, 800, "0 RPM", 150},
  {0, "Spoolup Rampup", createNumberField, 0x50+75, nil, 1, 20, "0 us"},
  {1, "Governor Response", createNumberField, 0x50+61, nil, 1, 10 },
  {1, "Holding Performance", createNumberField, 0x50+70, nil, 1, 10 },
  {1, "Holding Limit", createNumberField, 0x50+92, nil, 50, 100, "%" },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}
  
pageStabi = {
  -- { name, type, default, value, min, max }
  {0, "> STABI", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {1, "Function", createChoiceField, 0x50+30, nil, {{"Disabled", 65}, {"Rescue (Normal)", 66}, {"Rescue (Acro)", 67}, {"Rescue (Auto)", 72}, {"Stabilisation (Normal)", 68}, {"Stabilisation (Acro)", 69}, {"Stabilisation (Scale)", 70}, {"Coaxial", 71}, {"Selected by Channel", 73}}},
  {1, "Rescue Collective Pitch", createNumberField, 0x50+34, nil, 0, 100, "%" },
  {1, "Rescue Duration", createNumberField, 0x50+138, nil, 1, 250 },
  {1, "Hovering Collective Pitch", createNumberField, 0x50+139, nil, 0, 100, "%" },
  {1, "Stick Priority", createNumberField, 0x50+36, nil, 0, 16 },
  {1, "Direction Control Rate", createNumberField, 0x50+33, nil, 0, 5 },
  {1, "Acro Delay", createNumberField, 0x50+73, nil, 0, 30 },
  {1, "Rescue Auto - Duration", createNumberField, 0x50+38, nil, 1, 50, " s" },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageLimits = {
  -- { name, type, default, value, min, max }
  {0, "> LIMITS", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "CYCLIC RING", createTextLabel, nil, nil, "" },
  {0, "Aileron / Elevator Range", createNumberField, 0x50+10, nil, 32, 255 },
  {0, "", createTextLabel, nil, nil, "" },
  {0, "COLLECTIVE PITCH", createTextLabel, nil, nil, "" },
  {0, "Pitch Range", createNumberField, 0x50+12, nil, 32, 255 },
  {0, "", createTextLabel, nil, nil, "" },
  {0, "RUDDER END-POINTS", createTextLabel, nil, nil, "" },
  {0, "Minimum Limit", createNumberField, 0x50+14, nil, 32, 255 },
  {0, "Maximum Limit", createNumberField, 0x50+13, nil, 32, 255 },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageExpert = {
  -- { name, type, default, value, min, max }
  {0, "> EXPERT SETTINGS", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "Rotor Rotation Direction", createChoiceField, 0x50+41, nil, {{"Clockwise", 0}, {"Counter-Clockwise", 1}}},
  {1, "Stick Deadband", createNumberField, 0x50+28, nil, 2, 30 },
  {1, "Elevator Pitchup Comp.", createNumberField, 0x50+27, nil, 0, 4 },
  {0, "Cyclic Phase", createNumberField, 0x50+39, nil, -90, 90, nil, -127 },
  {1, "Pitchpump Booster", createNumberField, 0x50+63, nil, 0, 4 },
  {0, "Signal Processing", createChoiceField, 0x50+40, nil, {{"Normal", 0}, {"Extended", 1}}},
  {0, "Rotor Head", createChoiceField, 0x50+45, nil, {{"Flybarless", 48}, {"Flybar (Stabilized)", 49}, {"Flybar (Classic)", 50}}},
  {0, "RPM Sensor Filter", createNumberField, 0x50+74, nil, 1, 8 },
  {0, "Autorot. Bailout Rate", createNumberField, 0x50+84, nil, 1, 6, "s"},
  {0, "Stabi Correction", createChoiceField, 0x50+156, nil, {{"Off", 0}, {"On", 1}}},
  {1, "Anti-Gravity", createNumberField, 0x50+157, nil, 0, 100, "%" },
  {1, "Rescue Smoothing Factor", createNumberField, 0x50+159, nil, 0, 10 },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageAdvanced = {
  -- { name, type, default, value, min, max }
  {0, "> ADVANCED", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {0, "Cyclic", createTextLabel, nil, nil, "" },
  {0, "Geometry 6deg", createNumberField, 0x50+19, nil, 64, 250 },
  {1, "Feed Forward", createNumberField, 0x50+25, nil, 1, 12 },
  {1, "Elevator Filter", createNumberField, 0x50+42, nil, 0, 4 },
  {0, "Rudder", createTextLabel, nil, nil, "" },
  {1, "Delay", createNumberField, 0x50+43, nil, 0, 30 },
  {1, "Pirouette Consistency", createNumberField, 0x50+37, nil, 130, 250 },
  {1, "Dynamic", createNumberField, 0x50+29, nil, 3, 10, nil},
  {1, "Revomix", createNumberField, 0x50+32, nil, 0, 10, nil, -128 },
  {0, "", createTextLabel, nil, nil, "" },
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageLog = {
  -- { name, type, default, value, min, max }
  {0, "> LOG VIEWER", createTextLabel, nil, nil, "" },
  {0, "Duration", createNumberField, 0x50+171, nil, 0, 255, "0 seconds" },
  {0, "Previous flight", createChoiceField, 0x50+172, nil, {{"No", 0}, {"Yes", 1}}},
  {0, "", createMenuButton, nil, pagePrev, "< Back"},
}

pageMenu = {
  -- { name, type, default, value, min, max }
  {0, "SPIRIT FBL - Integration Menu", createTextLabel, nil, nil, "" },
  {1, "Bank", createChoiceField, 0x50+3, nil, {{"Bank 0", 0}, {"Bank 1", 1}, {"Bank 2", 2}}},
  {1, "", createMenuButton, nil, pageGeneral, "General"},
  {1, "", createMenuButton, nil, pageThrottle, "Throttle/Governor"},
  {1, "", createMenuButton, nil, pageLog, "Log Viewer"},
  {1, "", createMenuButton, nil, pageServos, "Servos"},
  {1, "", createMenuButton, nil, pageSensor, "Sensor"},
  {1, "", createMenuButton, nil, pageStabi, "Stabi"},
  {1, "", createMenuButton, nil, pageLimits, "Limits"},
  {1, "", createMenuButton, nil, pageAdvanced, "Advanced"},
  {1, "", createMenuButton, nil, pageExpert, "Expert"},
  {0, "", createTextLabel, nil, nil, "" },
  {1, "", createTextButton, 0x50+160, nil, "Save Settings"},
  {1, "", createTextButton, 0x50+161, nil, "Close Menu"},
}

local function updateField(index)
  if parameters[index][5] ~= nil then
    fields[index]:enable(true)
  end
end

local function create()
  requestInProgress = false
  requestLogInProgress = false
  refreshIndex = 0
  modifications = {}
  fields = {}
  pagePrev = pageMenu
  log_duration = 0
  log_loaded = 0
  skipField = 0

  local sensor = sport.getSensor(0x0C40);
  
  updateParameters(pageMenu)
  
  sensor:idle()

  return {sensor=sensor}
end

local function wakeup(widget)
  if requestExit then
    requestBankInProgress = false
    requestLogInProgress = false
    requestInProgress = false
    fields = {}
    modifications = {}
    logfields = {}
    logentries = {}
    pagePrev = pageMenu
    requestBank = false
    banksEnabled = false
    bankNumber = 0
    refreshIndex = 0
    skipField = 0
    log_duration = 0
    log_loaded = 0
    requestExit = false
    system.exit()
    return true
  end
  
  if requestBank then
    if requestBankInProgress then
      local value = widget.sensor:getParameter()
      if value ~= nil then
        local fieldId = value % 256
        
        if fieldId == 0x50+58 then
          value = math.floor(value / 256)
          
          if value == 7 then
            banksEnabled = false
          else
            banksEnabled = true
          end
          
          if value ~= nil then
            requestBank = false
            requestBankInProgress = false
          else
            banksEnabled = false
          end
        end
      else
        requestBankInProgress = false
        refreshIndex = 0
      end
    else
      if widget.sensor:requestParameter(0x50+58) then
        requestBankInProgress = true
      end
    end
  else
    if requestInProgress then
      local value = widget.sensor:getParameter()
      if value then
        local fieldId = value % 256
        local parameter = parameters[refreshIndex + 1]
        if fieldId == parameter[4] then
          value = math.floor(value / 256)

          if value ~= nil then
            if parameter[1] == 0 and bankNumber ~= 0 then
              parameters[refreshIndex + 1][5] = nil
              fields[refreshIndex + 1 - skipField]:enable(false)
            else
              parameters[refreshIndex + 1][5] = value
              fields[refreshIndex + 1 - skipField]:enable(true)
            end
          end
          
          -- disable Bank Switching
          if parameter[4] == 0x50+3 and banksEnabled == false then
            --skipField = 0
            fields[refreshIndex + 1 - skipField]:enable(false)
          end
          
          --if parameter[4] == 0x50+171 then
            --fields[refreshIndex + 1 - skipField]:enable(false)
          if parameter[4] == 0x50+172 then
            getLog()
            --fields[refreshIndex + 1 - skipField]:enable(false)
          end

          refreshIndex = refreshIndex + 1
          requestInProgress = false
        end
      else
        requestInProgress = false
      end
    elseif requestLogInProgress then
      local value = widget.sensor:getParameter()
      if value then
        local fieldId = value % 256
        if fieldId == 0x50+173 then
          value = math.floor(value / 256)
          if value ~= nil and value < 0x8000 then            
            local string = ""
            
            local LOG_EVENT_OK = 0x0
            local LOG_EVENT_OSC = 0x1
            local LOG_EVENT_CYCRING = 0x2
            local LOG_EVENT_RUDLIM = 0x4
            local LOG_EVENT_VIBES = 0x8
            local LOG_EVENT_HANG = 0x10
            local LOG_EVENT_RXLOSS = 0x20
            local LOG_EVENT_LOWVOLT = 0x40
            local LOG_EVENT_GOVENGAGED = 0x80
            local LOG_EVENT_RXCORRUPT = 0x100
            local LOG_EVENT_GOVSIGNAL = 0x200
            local LOG_EVENT_GOVLOST = 0x400
            local LOG_EVENT_GEOLINK = 0x800
            local LOG_EVENT_GEODRIFT = 0x1000
            local LOG_EVENT_GEODATA = 0x2000
            local LOG_EVENT_REBOOT = 0x4000
            local LOG_EVENT_ACCERR = 0x8000
            
            local log_message = { "Good Health Message",
              "Backup Oscillator is used",
              "Cyclic Ring Activated",
              "Rudder Limit Reached",
              "Vibration Level is very high",
              "Main Loop Hang Occured",
              "Receiver Signal Lost",
              "Power Voltage is low",
              "Governor was Engaged",
              "Calibration Finished",
              "Received Frame was Corrupted",
              "RPM Sensor data are too noisy",
              "RPM Sensor data are lost",
              "GeoLink connection is broken",
              "GeoLink function is imprecise",
              "GeoLink Data were Corrupted",
              "Data are not available" }
            
            local seconds = 0
            local minute = 0
            
            for index = 1, log_loaded do                 
              seconds = seconds + 1
              
              if seconds == 6 then
                minute = minute + 1
                seconds = 0
              end
            end

            if bitoper(value, LOG_EVENT_CYCRING) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[3])
            end
            if bitoper(value, LOG_EVENT_RUDLIM) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[4])
            end
            if bitoper(value, LOG_EVENT_VIBES) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[5])
            end
            if bitoper(value, LOG_EVENT_HANG) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0 [!]")
              form.addStaticText(line, nil, log_message[6])
            end
            if bitoper(value, LOG_EVENT_RXLOSS) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0 [!]")
              form.addStaticText(line, nil, log_message[7])
            end
            if bitoper(value, LOG_EVENT_LOWVOLT) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0 [!]")
              form.addStaticText(line, nil, log_message[8])
            end
            if bitoper(value, LOG_EVENT_GOVENGAGED) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[9])
            end
            if bitoper(value, LOG_EVENT_RXCORRUPT) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[11])
            end
            if bitoper(value, LOG_EVENT_GOVSIGNAL) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[12])
            end
            if bitoper(value, LOG_EVENT_GOVLOST) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0 [!]")
              form.addStaticText(line, nil, log_message[13])
            end  
            if bitoper(value, LOG_EVENT_GEOLINK) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0 [!]")
              form.addStaticText(line, nil, log_message[14])
            end  
            if bitoper(value, LOG_EVENT_GEODRIFT) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[15])
            end
            if bitoper(value, LOG_EVENT_GEODATA) > 0 then
              local line = form.addLine(minute .. ":" .. seconds .. "0")
              form.addStaticText(line, nil, log_message[16])
            end
          end

          log_loaded = log_loaded + 1
          
          if log_loaded < log_duration then
            lcd.setWindowTitle ("Loading (".. math.floor (log_loaded * 100 / log_duration) .. "%)")
          else
            lcd.setWindowTitle ("Loaded")
          end

          requestLogInProgress = false
        end
      else
        requestLogInProgress = false
      end
    else
      if #modifications > 0 then
        if widget.sensor:writeParameter(modifications[1][1], modifications[1][2]) == true then
          if modifications[1][1] == 0x50+160 then
            local buttons = {
              {label="OK", action=function() return true end},
            }
            form.openDialog("Spirit FBL", "Settings Saved successfully", buttons)
          end
          if modifications[1][1] == 0x50+161 then
            requestExit = true
          end
          refreshIndex = 0
          requestInProgress = false
          modifications[1] = nil
          skipField = 0
        end
      elseif refreshIndex < (#parameters) then
        local parameter = parameters[refreshIndex + 1]
	
        if parameter[4] ~= nil and parameter[4] ~= 0x50+160 and parameter[4] ~= 0x50+161 then
          if widget.sensor:requestParameter(parameter[4]) then -- requesting parameter values
            requestInProgress = true
          end
        else
          refreshIndex = refreshIndex + 1
          --skipField = skipField + 1
        end
        lcd.invalidate()
      elseif log_loaded < log_duration then -- flight log loading	
        if widget.sensor:requestParameter(0x50+173) then
          requestLogInProgress = true
        end
      end
    end
  end
end

local function event(widget, category, value, x, y)
  print("Event received:", category, value, x, y, KEY_EXIT_BREAK)
  if category == EVT_KEY and value == KEY_EXIT_BREAK then
    widget.sensor:idle(false)
  end
  return false
end

local icon = lcd.loadMask("spirit.png")

local function init()
  system.registerSystemTool({name=name, icon=icon, create=create, wakeup=wakeup, event=event})
end

return {init=init}
